/***************************************************************************
 * USB Servo controller
 * Andreas Goelzer
 * http://andreas.goelzer.de/usbservo
 ***************************************************************************
 * controls up to 12 servos on an attiny 2313
 * feel free to redistribute the source code under the terms of the gpl
 **************************************************************************/

#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
extern "C"{
#include "uart.h"
}

// uint16_t uart_geti(){
// 	uint16_t rv = 0;
// 	char read = 0;
// 	while(read <= ' ') read = uart_getc();
// 	while(read <= '9' && read >= '0'){
// 		rv *= 10;
// 		read = uart_getc();
// 	}
// 	return rv;
// }

void uart_puti(uint16_t out){
	char buffer[7];
	itoa(out,buffer,10);
	uart_puts(buffer);
}

struct machinestate {
	uint8_t portb, portd;
	uint16_t starttime;
	machinestate(uint16_t starttime_ = 0) : portb(0xff), portd(0x78), starttime(starttime_) {}
	inline void activate(){
		PORTB = portb;
		PORTD = portd;
	}
	void deactivateOutput(uint8_t number){
		if(number < 8){
			portb -= (1 << (7 - number));
		} else if(number == 8) portd -= (1 << PD6);
		else portd -= (1 << (number - 6));
	}
};

uint16_t servosettings[12];
machinestate machinestates[13];

uint8_t currentState = 0;
uint8_t stateCount = 0;

void convertServoValues(){

	uint16_t t = 0;
	machinestates[0] = machinestate();
	for(uint8_t j = 0; j < 12;  ++j){
		if(servosettings[j] == 0) machinestates[0].deactivateOutput(j);
	}
	
	for(currentState = 1; currentState < 13; ++currentState){
		uint16_t min = 65535;
		
		for(uint8_t j = 0; j < 12;  ++j){
			if(servosettings[j] > t && servosettings[j] < min) min = servosettings[j];
		}

		if(min == 65535){
			break;
		}
		machinestates[currentState] = machinestate(min);

		for(uint8_t j = 0; j < 12;  ++j){
			if(servosettings[j] <= min) machinestates[currentState].deactivateOutput(j);
		}
		t = min;
	}
	stateCount = currentState;
}

void introduceSelf(){
	uart_puts_P("\n\nusbservo version 2008-08-11\n");
	uart_puts_P("http://andreas.goelzer.de/usbservo\n\n");
}

void sGarbage(){
	uart_puts_P("garbage in s-command\n");
}

bool inputRead(){
	static char inputState = ' ';
	static char subState = 0;

	uint16_t c = uart_getc();
	if( c & UART_NO_DATA) return false;
	if(c & (UART_FRAME_ERROR | UART_FRAME_ERROR | UART_FRAME_ERROR)){
		uart_puts_P("UART err\n");
		return false;
	}
	uint8_t read = c;
	uart_putc(c);
	
	if(inputState == ' '){
		//drop whitespaces
		if(read <= ' ') return true;
		
		if(read == 'l'){
			uart_putc('\n');
			for(uint8_t j = 0; j < 12; ++j){
				uart_puti(j);
				uart_putc('\t');
				uart_puti(servosettings[j]);
				uart_putc('\n');
			}
			return true;
		}
		
		if(read == 'd'){
			for(uint8_t j = 0; j < stateCount; ++j){
				uart_puti(j);
				uart_putc('\t');
				uart_puti(machinestates[j].starttime);
				uart_putc('\t');
				uart_puti(machinestates[j].portb);
				uart_putc('\t');
				uart_puti(machinestates[j].portd);
				uart_putc('\n');
			}
			return true;
		}
		
		if(read == 's'){
			inputState = 's';
			subState = 0;
			return true;
		}
		
		if(read == 'v'){
			introduceSelf();
			return true;
		}
		
		return true;
		uart_puts_P("unknown command\n");
		
	}
	
	if(inputState == 's'){
		static uint8_t number;
		static uint16_t starttime;
		
		//before first parameter
		if(subState == 0){

			if(read <= ' ') return true;

			if('0' <= read && '9' >= read){
				number = read - '0';
				++subState;
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;
			
		}
		
		//in first parameter
		if(subState == 1){
			if(read <= ' '){
				++subState;
				return true;
			}

			if('0' <= read && '9' >= read){
				number = number * 10 + (read - '0');
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;

		}

		//before second parameter
		if(subState == 2){

			if(read <= ' ') return true;

			if('0' <= read && '9' >= read){
				starttime = read - '0';
				++subState;
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;
			
		}

		//in second parameter
		if(subState == 3){
			if(read <= ' '){
				if(number >= 12) uart_puts_P("only 12 channels\n");
				else{
					servosettings[number] = starttime;
					/*uart_puts_P("\nSet ");
					uart_puti(number);
					uart_puts_P(" to ");
					uart_puti(starttime);
					uart_putc('\n');*/
					convertServoValues();
				}
				subState = 0;
				inputState = ' ';
				return true;
			}

			if('0' <= read && '9' >= read){
				starttime = starttime * 10 + (read - '0');
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;

		}
		//uart_puts_P("\nunknown err");
		uart_puti(subState);
		uart_putc('\n');
		subState = 0;
		inputState = ' ';
		

	}

	inputState = ' ';
	return true;
}

int main(void){
	
	uart_init(UART_BAUD_SELECT(9600,F_CPU)); 
	TCCR1B = (1 << CS10);
	sei();
	uart_puts_P("\nProgram (re)start\n");

	introduceSelf();
	convertServoValues();

	for(;;){
		DDRD = machinestates[0].portd;
		DDRB = machinestates[0].portb;
		uint8_t nextState = 1;
		TCNT1 = 0;
		machinestates[0].activate();
		while(nextState < stateCount){
			if(TCNT1 > machinestates[nextState].starttime){
				machinestates[nextState].activate();
				++nextState;
			}
		}
		
		if(machinestates[stateCount - 1].starttime < TCNT1) while(TCNT1 < 64000) inputRead();
		TCNT1 = 0;
		while(TCNT1 < 64000) inputRead();
		TCNT1 = 0;
		while(TCNT1 < 32000) inputRead();
	}
   
    
}
