/***************************************************************************
 * USB Servo controller
 * Andreas Goelzer
 * http://andreas.goelzer.de/usbservo
 ***************************************************************************
 * controls up to 12 servos on an attiny 2313
 * feel free to redistribute the source code under the terms of the gpl
 **************************************************************************/

#include <stdlib.h>
#include <stdint.h>
#include <iostream>
#include <stdio.h>
void uart_puts(const char* s){
	std::cout << s;
}

void uart_puts_P(const char* s){
	std::cout << s;
}

void uart_putc(char c){
	std::cout << c;
}

char uart_getc(){
	
	return getchar();
}

int PORTB, PORTD;
const int PD6=6;
const int UART_NO_DATA = 0x0100;
// uint16_t uart_geti(){
// 	uint16_t rv = 0;
// 	char read = 0;
// 	while(read <= ' ') read = uart_getc();
// 	while(read <= '9' && read >= '0'){
// 		rv *= 10;
// 		read = uart_getc();
// 	}
// 	return rv;
// }

void uart_puti(uint16_t out){
	std::cout << out;

}

struct machinestate {
	uint8_t portb, portd;
	uint16_t starttime;
	machinestate(uint16_t starttime_ = 0) : portb(0xff), portd(0x78), starttime(starttime_) {}
	inline void activate(){
		PORTB = portb;
		PORTD = portd;
	}
	void deactivateOutput(uint8_t number){
		if(number < 8){
			portb -= (1 << (7 - number));
		} else if(number == 8) portd -= (1 << 6);
		else portd -= (1 << (number - 6));
	}
};

uint16_t servosettings[12];
machinestate machinestates[12];

uint8_t currentState = 0;
uint8_t stateCount = 0;

void convertServoValues(){

	uint16_t t = 0;
	for(currentState = 0; currentState < 12; ++currentState){
		uint16_t min = 65535;
		
		for(uint8_t j = 0; j < 12;  ++j){
			if(servosettings[j] > t && servosettings[j] < min) min = servosettings[j];
		}

		if(min == 65535){
			break;
		}
		machinestates[currentState] = machinestate(min);

		for(uint8_t j = 0; j < 12;  ++j){
			if(servosettings[j] <= min) machinestates[currentState].deactivateOutput(j);
		}
		t = min;
	}
	stateCount = currentState;
}

void introduceSelf(){
	uart_puts_P("\n\nusbservo version 2008-08-11\n");
	uart_puts_P("http://andreas.goelzer.de/usbservo\n\n");
}

void sGarbage(){
	uart_puts_P("garbage in s-command\n");
}

bool inputRead(){
	static char inputState = ' ';
	static char subState = 0;

	uint16_t c = uart_getc();
	if( c & UART_NO_DATA) return false;
	uint8_t read = c;
	
	if(inputState == ' '){
		//drop whitespaces
		if(read <= ' ') return true;
		
		if(read == 'l'){
			uart_putc('\n');
			for(uint8_t j = 0; j < 12; ++j){
				uart_puti(j);
				uart_putc('\t');
				uart_puti(servosettings[j]);
				uart_putc('\n');
			}
			return true;
		}
		
		if(read == 'd'){
			for(uint8_t j = 0; j < stateCount; ++j){
				uart_puti(j);
				uart_putc('\t');
				uart_puti(machinestates[j].starttime);
				uart_putc('\t');
				uart_puti(machinestates[j].portb);
				uart_putc('\t');
				uart_puti(machinestates[j].portd);
				uart_putc('\n');
			}
			return true;
		}
		
		if(read == 's'){
			inputState = 's';
			subState = 0;
			return true;
		}
		
		if(read == 'v'){
			introduceSelf();
			return true;
		}
		
		uart_puts_P("unknown command\n");
		return true;
		
	}
	
	if(inputState == 's'){
		static uint8_t number;
		static uint16_t starttime;
		
		//before first parameter
		if(subState == 0){

			if(read <= ' ') return true;

			if('0' <= read && '9' >= read){
				number = read - '0';
				++subState;
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;
			
		}
		
		//in first parameter
		if(subState == 1){
			if(read <= ' '){
				++subState;
				return true;
			}

			if('0' <= read && '9' >= read){
				number = number * 10 + (read - '0');
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;

		}

		//before second parameter
		if(subState == 2){

			if(read <= ' ') return true;

			if('0' <= read && '9' >= read){
				starttime = read - '0';
				++subState;
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;
			
		}

		//in second parameter
		if(subState == 3){
			if(read <= ' '){
				if(number >= 12) uart_puts_P("only 12 channels\n");
				else{
					servosettings[number] = starttime;
					uart_puts_P("\nSet ");
					uart_puti(number);
					uart_puts_P(" to ");
					uart_puti(starttime);
					uart_putc('\n');
					convertServoValues();
				}
				subState = 0;
				inputState = ' ';
				return true;
			}

			if('0' <= read && '9' >= read){
				starttime = starttime * 10 + (read - '0');
				return true;
			}
			
			sGarbage();
			subState = 0;
			inputState = ' ';
			return true;

		}
		uart_puts_P("\nunknown state ");
		uart_puti(subState);
		uart_putc('\n');
		subState = 0;
		inputState = ' ';
		

	}

	inputState = ' ';
	return true;
}

int main(void){
	
	uart_puts_P("\nProgram (re)start\n");
	introduceSelf();

	for(;;){
		inputRead();
	}
   
    
}
